/*
 * chk4mail - mail folder check
 * $Id: chk4mail.c,v 1.36 2002/09/20 20:00:20 johans Exp $
 */

/*
 * Copyright (C) 1998 - 2002, Johan van Selst <johans@stack.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#define __USE_BSD
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#undef __USE_BSD
#include <string.h>
#include <dirent.h>
#include <utime.h>
#include <err.h>
#include <sys/utsname.h>
#ifdef __linux__
#include <getopt.h>
#endif

#define MAXHEAD 256
#define MAXDIR 1024
#define MAXFILE 80

/* probably shouldn't have this bit hardcoded */
#define REMOTE_PATH "/usr/local/bin/chk4mail"

int			spc = 5, hideold = 0, mlen = 4, quiet = 0, brief = 0, oflag = 0,
			oanswer = 0, unreadasold = 0;
static char	rcsid[] = "$Id: chk4mail.c,v 1.36 2002/09/20 20:00:20 johans Exp $";

struct		summary
{
	int		new;
	int		flag;
	int		answer;
	int		total;
};

void
usage()
{
	/* give clue */
	fputs("chk4mail 2.12 - mail folder check\n"
		"Mbox, MMDF, MH and Maildir folders are recognized.\n\n"
		"-# <n>     Pad the number of messages with n spaces\n"
		"-a         Show how many messages have been answered\n"
		"-c         Include folder names with capitals\n"
		"-d <dir>   Look for mailfolders in this directory instead\n"
		"-f         Show how many messages have been flagged as important\n"
		"-h         Hide folders containing only old mail\n"
		"-i         Only parse standard incoming mailfolder\n"
		"-l         Always follow symbolic links\n"
		"-n         Only parse folders that received new mail\n"
		"-q         Suppress errors about missing or unreadable files\n"
		"-r <host>  Execute the command on an other host (with rsh)\n\n"
		"-s         Short output shows only files with new mail, implies -n\n"
		"-u         Don't treat unread messages as new mail\n"
		"Folders may be specified as arguments as well (filename or =folder)\n"
		"Read the manual for more details.\n",
		stderr);
	exit(1);
}

void
fail(const char *msg, const char *file)
{
	/* write friendly message */
	errx(1, "%s: %s\n", msg, file);
}

int
parse_mbox_mmdf(FILE *file, struct summary *sum, int mmdf)
{
	int nrs, isread, next, header = 1;
	long skip;
	char line[MAXHEAD+1];

	nrs = isread = 0;
	while (!feof(file))
	{
		skip = next = 0;

		/* parse headers */
		while (fgets(line, MAXHEAD, file) && strlen(line) > 1)
		{
			if ((!mmdf && !strncmp(line, "From ", 5)) ||
				(mmdf && !strcmp(line, "\1\1\1\1")))
				next = header = 1;
			else if (header)
			{
				if (strlen(line) < 2)
					header = 0;
				else if (!strncmp(line, "Content-Length: ", 16))
					skip = atol(line + 16);
				else if (!strncmp(line, "Status: RO", 10) ||
					 !strncmp(line, "Status: OR", 10))
					isread++;
				else if (unreadasold && !strncmp(line, "Status: O", 9))
					isread++;
				else if (!strncmp(line, "X-Status: ", 10))
				{
					if (strchr(line, 'F'))
						sum->flag++;
					if (strchr(line, 'A'))
						sum->answer++;
				}
			}
		}

		/* skip body (hopefully) */
		if (next)
		{
			nrs++;
			fseek(file, skip, SEEK_CUR);
		}
	}

	if (mmdf)
		nrs--;

	sum->total = nrs;
	sum->new = nrs - isread;

	return 0;
}

int
parse_mbox(FILE *file, struct summary *sum)
{
	return parse_mbox_mmdf(file, sum, 0);
}

int
parse_mmdf(FILE *file, struct summary *sum)
{
	return parse_mbox_mmdf(file, sum, 1);
}

int
parse_maildir(const char *subdir, struct summary *sum)
{
	char			*filename, *p;
	struct dirent	*file;
	DIR				*dir;
	int				isread;

	sum->new = 0;
	sum->total = 0;

	asprintf(&filename, "%s/cur", subdir);
	if (!(dir = opendir(filename)))
		fail("Unable to open subdirectory", filename);

	while ((file = readdir(dir)))
	{
		if (!strcmp(file->d_name, ".") || !strcmp(file->d_name, ".."))
			continue;
		isread = 0;
		if ((p = strrchr(file->d_name, ':')) && !strncmp(p, ":2,", 3))
		{
			p += 2;
			while (*(++p))
				if (*p == 'S')
					isread = 1;
				else if (oanswer && *p == 'R')
					(sum->answer)++;
				else if (oflag && *p == 'F')
					(sum->flag)++;
		}
		if (isread || unreadasold)
			(sum->total)++;
		else
			(sum->new)++;
	}
	closedir(dir);

	sprintf(filename, "%s/new", subdir);
	if (!(dir = opendir(filename)))
		fail("Unable to open subdirectory", filename);

	while ((file = readdir(dir)))
		if (strcmp(file->d_name, ".") && strcmp(file->d_name, ".."))
			(sum->new)++;
	closedir(dir);

	sum->total += sum->new;
	free(filename);
	return 0;
}

int
parse_mh(const char *subdir, struct summary *sum)
{
	DIR				*dir;
	struct dirent	*file;
	struct stat		statbuf;
	const int		len = strlen(subdir) + 7;
	char			*x, *filename;
	char			real;

	if (!(dir = opendir(subdir)))
		fail("Unable to open subdirectory", subdir);

	filename = malloc(len);
	memset(filename, 0, len);
	/* total = #filenames consisting of numbers */
	for (sum->total = 0; (file = readdir(dir)); )
	{
		for (real = 1, x = file->d_name; *x; x++)
			real &= *x >= '0' && *x <= '9';
		if (real)
		{
			(sum->total)++;
			snprintf(filename, len-1, "%s/%s", subdir, file->d_name);
			if (!stat(filename, &statbuf))
				if (statbuf.st_mtime >= statbuf.st_atime)
					(sum->new)++;
		}
	}

	closedir(dir);
	free(filename);
	return 0;
}

int
parse_dir(const char *dir, struct summary *sum)
{
	struct stat		filestat;
	char			*filename;

	asprintf(&filename, "%s/cur", dir);
	if (lstat(filename, &filestat) < 0 || !S_ISDIR(filestat.st_mode))
	{
		/* parse MH-style mailfolder */
		free(filename);
		return parse_mh(dir, sum);
	}
	else
	{
		/* parse cur/new/tmp-style mailfolder */
		free(filename);
		return parse_maildir(dir, sum);
	}
}

int
desc(char *buf, const int i, const char *msg)
{
	if (i <= 0)
		return sprintf(buf, "%*s %s", spc, i ? "?" : "-", msg);
	else
		return sprintf(buf, "%*d %s", spc, i, msg);
}

int
display(const char *folder, struct summary *sum)
{
	static char	line[256];
	char		*p = line;

	if ((sum->total > 0) && (sum->new || !hideold))
	{
		p += sprintf(p, "%*s", mlen, folder);
		p += desc(p, sum->new, "new");
		if (oflag)
			p += desc(p, sum->flag, "flag");
		if (oanswer)
			p += desc(p, sum->answer, "answ");
		fprintf(stdout, "%s of total%*d\n", line, spc, sum->total);
	}
	else if (brief && sum->new > 0)
		fprintf(stdout, "%s\n", folder);

	return 0;
}

int
parse_folder(const char *dir, const char *filenm, int flink)
{
	struct stat		filestat;
	struct utimbuf	filetime;
	struct summary	sum = { 0, 0, 0, 0 };
	char			*filename;
	FILE			*file;
	char			c;
	int				new, total;

	asprintf(&filename, "%s/%s", dir, filenm);

	if (lstat(filename, &filestat) < 0)
	{
		if (quiet)
			return 2;
		else
			fail("Unable to stat file", filename);
	}

	/* Follow symlinks iff 'flink' */
	if (S_ISLNK(filestat.st_mode) && flink &&
		(stat(filename, &filestat) < 0))
	{
		if (quiet)
			return 2;
		else
			fail("Unable to stat file", filename);
	}
	if (S_ISDIR(filestat.st_mode))
	{
#ifdef	SUBMAILDIR
		char *subsub = malloc(MAXFILE), *p;
		DIR *subdir;
		struct dirent *subfile;
#endif	/* SUBMAILDIR */

		/* parse directory mailfolder */
		new = total = 0;
		parse_dir(filename, &sum);
		display(filenm, &sum);

#ifdef	SUBMAILDIR
		if (!(subdir = opendir(filename)))
			fail("Unable to open subdirectory", filename);

		/* find maildir subdirectories */
		while ((subfile = readdir(subdir)))
		{
			if (!strcmp(subfile->d_name, ".") ||
					!strcmp(subfile->d_name, ".."))
				continue;
			/* subdirs 'new' and 'cur' should exist - prolly 'tmp' too */
			snprintf(subsub, MAXFILE-1, "%s/%s/new", filename, subfile->d_name);
			if ((stat(subsub, &filestat) < 0) || !S_ISDIR(filestat.st_mode))
				continue;
			snprintf(subsub, MAXFILE-1, "%s/%s/cur", filename, subfile->d_name);
			if ((stat(subsub, &filestat) < 0) || !S_ISDIR(filestat.st_mode))
				continue;
			/* looks like a maildir, let's check it out... */
			if (!(p = strrchr(subsub, '/')))
				continue;	/* silently ignore this error */
			*p = '\0';
			parse_maildir(subsub, &sum);

			snprintf(subsub, MAXFILE-1, "../%s", subfile->d_name);
			display(subsub, &sum);
		}
#endif	/* SUBMAILDIR */

		free(subsub);
		free(filename);
		return 1;
	}
	if (!S_ISREG(filestat.st_mode))
		return 2;
	if (!filestat.st_size)
		return 3;

	/* open and parse mailfolder */
	filetime.actime = filestat.st_atime;
	filetime.modtime = filestat.st_mtime;

	if (hideold > 1)
	{
		if (filetime.modtime < filetime.actime)
		{
			free(filename);
			return 0;
		}
		else if (brief)
		{
#if 0
			display(filenm, 1, 0);
#endif
			free(filename);
			return 0;
		}
	}

	if (!(file = fopen(filename, "r")))
	{
		if (quiet)
			return 4;
		else
			fail("Unable to open file", filename);
	}

	if ((c = fgetc(file)) == '\1')
	{
		ungetc(c, file);
		parse_mmdf(file, &sum);
	}
	else
	{
		ungetc(c, file);
		parse_mbox(file, &sum);
	}
	display(filenm, &sum);

	/* close and preserve timestamp */
	fclose(file);
	utime(filename, &filetime);
	free(filename);

	return 0;
}

int
main(int argc, char **argv)
{
	char *maildir = NULL;
	char *inbox = NULL;
	char *remote = NULL;
	DIR *dir;
	struct dirent *file;
	char *slash;
	char dirlist[MAXDIR][MAXFILE];
	struct utsname uts;
	int opt, max, len, inonly = 0, flink = 0, caps = 0;

	len = strlen(argv[0]);
	if ((len > 5 && !strcmp(argv[0] + len - 5, "/nfrm")) ||
		(len == 4 && !strcmp(argv[0], "nfrm")))
		inonly = 1;

	while ((opt = getopt(argc, argv, "#:acd:fhilnqr:su")) != EOF)
		switch (opt)
		{
		case '#':
			if ((spc = atoi(optarg)) <= 0)
				usage();
			break;
		case 'a':
			oanswer = 1;
			break;
		case 'c':
			caps = 1;
			break;
		case 'd':
			maildir = strdup(optarg);
			break;
		case 'f':
			oflag = 1;
			break;
		case 'h':
			hideold = 1;
			break;
		case 'i':
			inonly = 1;
			break;
		case 'l':
			flink = 1;
			break;
		case 'n':
			hideold = 2;
			break;
		case 'q':
			quiet = 1;
			break;
		case 'r':
			remote = strdup(optarg);
			break;
		case 's':
			brief = 1;
			break;
		case 'u':
			unreadasold = 1;
			break;
		default:
			usage();
		}

	if (brief)
		hideold = 2;
	if (remote)
	{
		if (uname(&uts) < 0)
			fail("uname() failed", "???");
		if ((slash = strchr(uts.nodename, '.')))
			slash[0] = '\0';
		if (strcmp(remote, uts.nodename))
		{
			while (--argc)
				*(argv[argc]-1) = ' ';
			/* execl() does not return on success */
			if (execl("/usr/bin/rsh", remote, *argv, NULL) < 0)
				fail("execl() failed", remote);
		}
	}
	/* find inbox */
	if ((inbox = getenv("MAIL")))
		inbox = strdup(inbox);

	/* Shortcut to the end */
	if (inonly)
		goto INBOX;

	/* find maildir */
	if (!maildir)
	{
		asprintf(&maildir, "%s/Mail", getenv("HOME"));
		if (!(dir = opendir(maildir)))
			sprintf(maildir, "%s/mail", getenv("HOME"));
	}
	if (!(dir = opendir(maildir)))
		fail("Unable to open maildirectory", maildir);

	/* only parse specified mailboxes and exit */
	if (optind < argc)
	{
		/* for a nice layout: find max filename length first */
		opt = optind;
		while (opt < argc)
		{
			if ((slash = strrchr(argv[opt], '/')))
				len = strlen(slash + 1);
			else
				len = strlen(argv[opt]);
			opt++;
			if (len > mlen)
				mlen = len;
		}
		opt = optind;
		while (opt < argc)
		{
			if (argv[opt][0] == '=' || argv[opt][0] == '+')
				parse_folder(maildir, argv[opt] + 1, 1);
			else if ((slash = strrchr(argv[opt], '/')))
			{
				slash[0] = '\0';
				parse_folder(argv[opt], slash+1, 1);
			}
			else
			{
				getcwd(maildir, MAXFILE);
				parse_folder(maildir, argv[opt], 1);
			}
			opt++;
		}
		closedir(dir);
		return 0;
	}

	/* sort maildir */
	max = 0;
	memset(dirlist, 0, MAXFILE*MAXDIR);
	while ((file = readdir(dir)))
	{
		/* check if this mailfolder is worth reading */
		if ((len = strlen(file->d_name)) &&
			(file->d_name[0] < 'a' || file->d_name[0] > 'z') &&
			(file->d_name[0] < 'A' || file->d_name[0] > 'Z' || !caps) &&
			(file->d_name[0] < '0' || file->d_name[0] > '9'))
			continue;

		if ((len > 4) && !strcmp((char *)(file->d_name + len-5), ".lock"))
			continue;

		if (len > mlen)
			mlen = len;
		strncpy(dirlist[max++], file->d_name, MAXFILE-1);
	}

	qsort(dirlist, MAXDIR, MAXFILE,
		(int (*)(const void *, const void *))strcmp);
	max = MAXDIR-max;

	/* parse maildir */
	while (max < MAXDIR)
		parse_folder(maildir, dirlist[max++], flink);
	closedir(dir);

INBOX:
	/* parse inbox */
	if (inbox && strlen(inbox))
	{
		slash = strrchr(inbox, '/');
		slash[0] = '\0';
		parse_folder(inbox, slash+1, 1);
	}

	return 0;
	/* It's no unused variable... See? Here it is... */
	(void)rcsid;
}
